from logging import getLogger
from gettext import gettext as _

from django.db import models

from OpenSSL import crypto as openssl

from pulpcore.plugin import storage
from pulpcore.plugin.models import ContentGuard


log = getLogger(__name__)


class X509CertGuard(ContentGuard):
    """
    A content-guard that authenticates the request based on a client provided X.509 Certificate.

    Fields:
        ca_certificate (models.FileField): The CA certificate used to
            validate the client certificate.
    """

    TYPE = 'x509'

    def _certpath(self, name):
        return storage.get_tls_path(self, name)

    ca_certificate = models.FileField(max_length=255, upload_to=_certpath)

    def permit(self, request):
        """Authorize the specified web request.

        Args:
            request (aiohttp.web.Request): A request for a published file.

        Raises:
            PermissionError: When the request cannot be authorized.

        """
        ca = self.ca_certificate.read()
        validator = Validator(ca.decode('utf8'))
        validator(request)


class Validator:
    """An X.509 certificate validator."""

    SSL_CERTIFICATE_HEADER = 'SSL-CLIENT-CERTIFICATE'

    @staticmethod
    def format(pem):
        """Ensure the PEM encoded certificate is properly formatted.

        The certificate is passed as an HTTP header which does not permit newlines.

        Args:
            pem (str): A PEM encoded certificate.

        Returns:
            str: A properly PEM formatted certificate.

        """
        header = '-----BEGIN CERTIFICATE-----'
        footer = '-----END CERTIFICATE-----'
        body = pem.replace(header, '')
        body = body.replace(footer, '')
        body = body.strip(' \n\r')
        return '\n'.join((header, body, footer))

    @staticmethod
    def load(pem):
        """Load the PEM encoded certificate.

        Encapsulates complexity of OpenSSL.

        Args:
            pem (str): A PEM encoded certificate.

        Returns:
            openssl.X509: The loaded certificate.

        Raises:
            ValueError: On load failed.

        """
        try:
            return openssl.load_certificate(openssl.FILETYPE_PEM, buffer=Validator.format(pem))
        except Exception as le:
            raise ValueError(str(le))

    def client_certificate(self, request):
        """Extract and load the client certificate passed in the SSL-CLIENT-CERTIFICATE header.

        Args:
            request (aiohttp.web.Request): A request for a published file.

        Returns:
            openssl.X509: The loaded certificate.

        Raises:
            KeyError: When the client certificate header has not
                been passed in the request.

        """
        name = self.SSL_CERTIFICATE_HEADER
        try:
            certificate = request.headers[name]
        except KeyError:
            reason = _('HTTP header "{h}" not found.').format(h=name)
            raise KeyError(reason)
        else:
            return Validator.load(certificate)

    def __init__(self, ca_certificate):
        """Inits a new validator.

        Args:
            ca_certificate (str): A PEM encoded CA certificate.

        """
        self.ca_certificate = self.load(ca_certificate)

    @property
    def store(self):
        """A X.509 certificate (trust) store.

        Returns:
            openssl.X509Store: A store containing the CA certificate.

        """
        store = openssl.X509Store()
        store.add_cert(self.ca_certificate)
        return store

    def __call__(self, request):
        """Validate the client X.509 certificate passed in the request.

        Args:
            request (aiohttp.web.Request): A request for a published file.

        Raises:
            PermissionError: When validation the client certificate
                cannot be validated.

        """
        try:
            context = openssl.X509StoreContext(
                certificate=self.client_certificate(request),
                store=self.store,
            )
            context.verify_certificate()
        except openssl.X509StoreContextError:
            raise PermissionError(_('Client certificate cannot be validated.'))
        except Exception as pe:
            raise PermissionError(str(pe))



"""Constants for Pulp certguard plugin tests."""
import os
from urllib.parse import urljoin

from pulp_smash.pulp3.constants import BASE_PATH, BASE_CONTENT_GUARDS_PATH  # noqa:F401
from pulp_file.tests.functional.constants import FILE_REMOTE_PATH, FILE_PUBLISHER_PATH  # noqa:F401

_CURRENT_DIR = os.path.dirname(os.path.realpath(__file__))
CERTS_BASE_PATH = os.path.join(
    _CURRENT_DIR, 'artifacts', 'x509', 'certificates'
)
CERT_CA_FILE_PATH = os.path.join(CERTS_BASE_PATH, 'ca.pem')
CERT_CLIENT_FILE_PATH = os.path.join(CERTS_BASE_PATH, 'client.pem')
KEYS_BASE_PATH = os.path.join(_CURRENT_DIR, 'x509', 'keys')
X509_CONTENT_GUARD_PATH = urljoin(BASE_CONTENT_GUARDS_PATH, "certguard/x509/")



from gettext import gettext as _

from rest_framework import serializers

from pulpcore.plugin.serializers import ContentGuardSerializer

from .models import X509CertGuard, Validator


class CertGuardSerializer(ContentGuardSerializer):
    """Content Guard Serializer."""

    ca_certificate = serializers.FileField(
        help_text="The Certificate Authority (CA) certificate.",
        write_only=True
    )

    class Meta:
        model = X509CertGuard
        fields = ContentGuardSerializer.Meta.fields + (
            'ca_certificate',
        )

    @staticmethod
    def validate_ca_certificate(certificate):
        """Validates the given certificate."""
        buffer = certificate.read()
        try:
            Validator.load(buffer.decode('utf8'))
        except ValueError:
            reason = _('Must be PEM encoded X.509 certificate.')
            raise serializers.ValidationError(reason)
        else:
            return certificate



from pulpcore.plugin.viewsets import ContentGuardFilter, ContentGuardViewSet

from .models import X509CertGuard
from .serializers import CertGuardSerializer


class CertGuardViewSet(ContentGuardViewSet):
    """X509CertGuard API Viewsets."""

    endpoint_name = 'x509'
    queryset = X509CertGuard.objects.all()
    serializer_class = CertGuardSerializer
    filterset_class = ContentGuardFilter



import unittest
from random import choice
from requests import HTTPError
from pulp_smash import api, config, utils
from pulp_smash.pulp3.constants import (
    DISTRIBUTION_PATH,
    REPO_PATH,
)
from pulp_smash.pulp3.utils import (
    download_content_unit,
    gen_distribution,
    gen_repo,
    publish,
    sync,
)
from pulp_certguard.tests.functional.constants import (
    CERT_CA_FILE_PATH,
    CERT_CLIENT_FILE_PATH,
    X509_CONTENT_GUARD_PATH,
    FILE_REMOTE_PATH,
    FILE_PUBLISHER_PATH
)
from pulp_certguard.tests.functional.utils import (
    gen_file_publisher,
    gen_file_remote,
    get_file_content_paths
)


class CertGuardTestCase(unittest.TestCase):
    """Api tests for cerguard."""

    @classmethod
    def setUpClass(cls):
        """Initialize common use entities and variables.

        0. Prepare artifacts (ca and client certs)
        1. Create self.contentguard for ca.pem located in CERT_CA_PATH.
        2. create a repo.
        3. create a file remote.
        4. Sync it.
        5. Create a Publisher.
        6. Create a Publication.

        """
        cls.cfg = cfg = config.get_config()
        cls.client = client = api.Client(cfg, api.page_handler)
        cls.teardown_cleanups = []

        # The certificate header must have no \n chars
        with open(CERT_CLIENT_FILE_PATH, 'r') as cert_client_file:
            cls.client_cert = str(cert_client_file.read()).replace('\n', '')

        with utils.ensure_teardownclass(cls):
            # 1. Create certguard
            with open(CERT_CA_FILE_PATH, 'rb') as cert_ca_file:
                cls.certguard = client.post(
                    X509_CONTENT_GUARD_PATH,
                    data={'name': utils.uuid4()},
                    files={'ca_certificate': cert_ca_file}
                )
                cls.teardown_cleanups.append(
                    (client.delete, cls.certguard['_href'])
                )

            # 2. Create a repo
            _repo = client.post(REPO_PATH, gen_repo())
            cls.teardown_cleanups.append((cls.client.delete, _repo['_href']))

            # 3. Create a remote
            cls.remote = client.post(FILE_REMOTE_PATH, gen_file_remote())
            cls.teardown_cleanups.append(
                (cls.client.delete, cls.remote['_href'])
            )

            # 4. Sync and read synced repo
            sync(cfg, cls.remote, _repo)
            cls.repo = client.get(_repo['_href'])

            # 5. Create a Publisher
            cls.publisher = client.post(
                FILE_PUBLISHER_PATH, gen_file_publisher()
            )
            cls.teardown_cleanups.append(
                (cls.client.delete, cls.publisher['_href'])
            )

            # 6. Create a publication
            cls.publication = publish(cfg, cls.publisher, cls.repo)
            cls.teardown_cleanups.append(
                (cls.client.delete, cls.publication['_href'])
            )

    def test_negative_download_protected_content_without_keys(self):
        """Assert content protected by cert-guard cannot be downloaded.

        1. Create a protected distribution using the self.contentguard.
        2. Assert content cannot be downloaded without cert and key.
        """
        # Create a protected distribution
        distribution = self.client.using_handler(api.task_handler).post(
            DISTRIBUTION_PATH,
            gen_distribution(
                publication=self.publication['_href'],
                content_guard=self.certguard['_href']
            )
        )
        self.addCleanup(self.client.delete, distribution['_href'])

        # Pick a filename
        unit_path = choice(get_file_content_paths(self.repo))

        # Try to download it without the SSL-CLIENT-CERTIFICATE
        with self.assertRaises(HTTPError):
            download_content_unit(self.cfg, distribution, unit_path)

    def test_positive_download_protected_content_with_keys(self):
        """Assert content protected by cert-guard can be downloaded.

        1. Create a protected distribution using the self.contentguard.
        2. Assert content can be downloaded using the proper cert and key.
        """
        # Create a protected distribution
        distribution = self.client.using_handler(api.task_handler).post(
            DISTRIBUTION_PATH,
            gen_distribution(
                publication=self.publication['_href'],
                content_guard=self.certguard['_href']
            )
        )
        self.addCleanup(self.client.delete, distribution['_href'])

        # Pick a filename
        unit_path = choice(get_file_content_paths(self.repo))

        # Try to download it passing the proper SSL-CLIENT-CERTIFICATE
        download_content_unit(
            self.cfg,
            distribution,
            unit_path,
            headers={'SSL-CLIENT-CERTIFICATE': self.client_cert}
        )

    def test_positive_add_contentguard_to_existing_distribution(self):
        """Assert adding contentguard to existing distribution works well.

        1. Create a distribution without protection
        2. Assert content can be downloaded
        3. Add contentguard to the distribution
        4. Assert content cannot be downloaded without key
        5. Assert content can be downloaded with key
        """
        # 1 unprotected distribution
        distribution = self.client.using_handler(api.task_handler).post(
            DISTRIBUTION_PATH,
            gen_distribution(publication=self.publication['_href'])
        )
        self.addCleanup(self.client.delete, distribution['_href'])

        # Pick a filename
        unit_path = choice(get_file_content_paths(self.repo))

        # Download it without certificate
        download_content_unit(self.cfg, distribution, unit_path)

        # Update distribution adding the guard
        distribution = self.client.using_handler(api.task_handler).patch(
            distribution['_href'],
            {'content_guard': self.certguard['_href']}
        )

        # Cannot download without key
        with self.assertRaises(HTTPError):
            download_content_unit(self.cfg, distribution, unit_path)

        # Try to download it passing the proper SSL-CLIENT-CERTIFICATE
        download_content_unit(
            self.cfg,
            distribution,
            unit_path,
            headers={'SSL-CLIENT-CERTIFICATE': self.client_cert}
        )

    def test_positive_remove_contentguard(self):
        """Assert that content can be download without guard if it is removed.

        1. Create a protected distribution using the self.contentguard
        2. Assert content cannot be downloaded without keys
        3. Remove the contentguard
        4. Assert content can be downloaded without keys
        """
        # Create a protected distribution
        distribution = self.client.using_handler(api.task_handler).post(
            DISTRIBUTION_PATH,
            gen_distribution(
                publication=self.publication['_href'],
                content_guard=self.certguard['_href']
            )
        )
        self.addCleanup(self.client.delete, distribution['_href'])

        # Pick a filename
        unit_path = choice(get_file_content_paths(self.repo))

        # Try to download it without the SSL-CLIENT-CERTIFICATE
        with self.assertRaises(HTTPError):
            download_content_unit(self.cfg, distribution, unit_path)

        # Update distribution removing the guard
        distribution = self.client.using_handler(api.task_handler).patch(
            distribution['_href'],
            {'content_guard': None}
        )

        # Now content can be downloaded
        download_content_unit(self.cfg, distribution, unit_path)

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        for cleanup_function, args in reversed(cls.teardown_cleanups):
            cleanup_function(args)
